import time 
import numpy as np
import torch 
torch.manual_seed(42)
torch.set_num_threads(1)

from utils import trace

def renormalize(T, Ux, Uy):
    chi, chi_new = Ux.shape[1], Ux.shape[2]
    UTT = torch.einsum('abc,adeg,befh->cdfgh',(Ux, T, T)).contiguous().view(chi_new,chi,chi,chi**2)
    UTT2 = torch.einsum('abcd,gefd->abecfg', (UTT, UTT)).contiguous().view(chi_new, chi**2, chi**2, chi_new)
    Uy = Uy.view(chi**2, chi_new)

    T = torch.einsum('abcd,be,cf->aefd', (UTT2, Uy, Uy)).contiguous()
    #T = torch.einsum('aem,bhn,fko,jlp,abcd,ecfg,gikl,dhij->mnop', (U, U, U, U, T, T, T, T))
    return T

class Ising(torch.nn.Module):
    def __init__(self, chi, niter, dtype=torch.float64, device='cpu', use_checkpoint=False):
        super(Ising, self).__init__()
    
        self.D = 2
        self.chi = chi # cutoff 
        self.niter = niter
        self.dtype = dtype
        self.device = device

        self.params = torch.nn.ParameterList()
        chi = self.D
        for step in range(niter): 
            chi_new = min(chi**2, self.chi)
            for _ in range(2): #xy direction 
                U = torch.randn(chi, chi, chi_new, dtype=dtype, device=device)
                self.params.append(torch.nn.Parameter(U)) 
            chi = min(chi**2, self.chi)

    def get_isometry(self, step):
        return self.params[2*step], self.params[step*2+1]

    def set_isometry(self, step, U):
        self.params[step].data = U

    def forward(self, T):
        lnZ = 0.0
        for n in range(self.niter):
            f = T.abs().max()
            lnZ += 4**(-n)*torch.log(f)
            T = T / f

            Ux, Uy = self.get_isometry(n) # isometry for this step
            T = renormalize(T, Ux, Uy)
           
        lnZ += torch.log(trace(T))/4.0**args.niter
        return lnZ

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=0.44068679350977147, help="beta")
    parser.add_argument("-chi", type=int, default=10, help="chi")
    parser.add_argument("-niter", type=int, default=20, help="niter")
    parser.add_argument("-nsweeps", type=int, default=10, help="nsweeps")
    parser.add_argument("-nstays", type=int, default=1, help="nstays")
    parser.add_argument("-exact", type=float, default=-0.9296953983414709, help="exact free energy")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)

    # initial tensor
    K = torch.tensor([args.beta], dtype=dtype, device=device)
    c = torch.sqrt(torch.cosh(K)/2.)
    s = torch.sqrt(torch.sinh(K)/2.)
    M = torch.stack([torch.cat([c+s, c-s]), torch.cat([c-s, c+s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M)) 

    model = Ising(args.chi, args.niter, dtype=dtype, device=device)
    
    def turn_on_grad(i):
        for j, p in enumerate(model.parameters()):
            if (j==i):
                p.requires_grad = True
            else:
                p.requires_grad = False

    sweeps = 0; step = 0; direction = 1
    for _ in range(2*args.nsweeps*args.niter):
        #sweep back and forth through isometries
        turn_on_grad(step)

        for s in range(args.nstays): # local sweeps
            model.zero_grad()
            start = time.time()
            lnZ = model.forward(T) 
            forward = time.time() 
            lnZ.backward()
            backward = time.time()
            
            with torch.no_grad(): 
                #compute environment from gradient  
                #\partial tr(wE)/ \partial w = E^T 
                #cf http://www.matrixcalculus.org
                E = model.params[step].grad
                chi, chi_new = E.shape[0], E.shape[-1]
                E = E.reshape(chi**2, chi_new)
                #and perform mera update
                U, S, V = torch.svd(E)
                svd = time.time()
                
                model.params[step].data = (U@V.t()).view(chi, chi, chi_new)

        #print(forward-start, backward - forward, svd-backward)
        step += direction

        #when reaching the end reverse sweep direction and report results
        if (step==2*args.niter-1 or step==0):
            direction *= -1
            sweeps += 1
            with torch.no_grad():
                lnZ = model.forward(T) 
            
            if args.exact is not None:
                print (sweeps, -lnZ.item(), (-lnZ.item()-args.exact)/np.abs(args.exact) )
            else:
                print (sweeps, -lnZ.item())
