import time 
import numpy as np
import torch 
torch.manual_seed(42)
torch.set_num_threads(1)
from torch.utils.checkpoint import checkpoint

from tensornets import CTMRG

def symmetrize(A):
    '''
    A(phy, up, left, down, right, phy')
    left-right, up-down, diagonal symmetrize
    '''
    Asymm = (A + A.permute(0, 1, 4, 3, 2, 5))/2.           # left-right symmetry
    Asymm = (Asymm + Asymm.permute(0, 3, 2, 1, 4, 5))/2.   # up-down symmetry
    Asymm = (Asymm + Asymm.permute(0, 4, 3, 2, 1, 5))/2.   # skew-diagonal symmetry
    Asymm = (Asymm + Asymm.permute(0, 2, 1, 4, 3, 5))/2.   # diagonal symmetry
 
    return Asymm

def ctmrg(T, chi, nctmrg):

    C, E = CTMRG(T, chi, nctmrg, use_checkpoint=args.use_checkpoint)
    C = C.detach()
    E = E.detach() 

    Z1 = torch.einsum('ab,bcd,fd,gha,chij,fjk,lg,mil,mk', (C,E,C,E,T,E,C,E,C))
    #CEC = torch.einsum('da,ebd,ce->abc', (C,E,C)).view(1, D**2*d)
    #ETE = torch.einsum('abc,bldr,mdn->almcrn',(E,T,E)).contiguous().view(D**2*d, D**2*d)
    #ETE = (ETE+ETE.t())/2.
    #Z1 = CEC@ETE@CEC.t()

    Z3 = torch.einsum('ab,bc,cd,da', (C,C,C,C))
    Z2 = torch.einsum('ab,bcd,de,fa,gcf,ge',(C,E,C,C,E,C))
    lnZ = torch.log(Z1) + torch.log(Z3) - 2.*torch.log(Z2)

    return lnZ

def renormalize(T, U):
    UTU1 = torch.einsum('blu,xabcdy,dmv->luxacymv', (U, T, U)).contiguous()
    UTU2 = torch.einsum('cdp,ybldmz,abq->cpylmzaq', (U, T, U)).contiguous()
    T = torch.einsum('luxacymv,cpylmzaq->xqupvz', (UTU1, UTU2)).contiguous()
    return T

class Ising(torch.nn.Module):
    def __init__(self, D, chi, nxtrg, nctmrg, dtype=torch.float64, device='cpu', use_checkpoint=False):
        super(Ising, self).__init__()
    
        self.D = D     # for XTRG isometry
        self.chi = chi # for CTMRG boundary dimension
        self.nxtrg = nxtrg 
        self.nctmrg = nctmrg 
        self.dtype = dtype
        self.device = device

        self.params = torch.nn.ParameterList()
        d = 2
        for step in range(nxtrg): 
            d_new = min(d**2, self.D)
            self.params.append(torch.nn.Parameter(torch.randn(d, d, d_new, dtype=dtype, device=device))) 
            d = min(d**2, self.D)

    def get_isometry(self, step):
        return self.params[step]

    def forward(self, T):
        lnZ = 0.0
        for n in range(self.nxtrg):
            f = T.abs().max()
            lnZ += torch.log(f)/2.0**n
            T = T / f

            U = self.get_isometry(n) # isometry for this step

            tensors = T, U
            if args.use_checkpoint: # use checkpoint to save memory
                T = checkpoint(renormalize, *tensors)
            else:
                T = renormalize(*tensors)

            T = symmetrize(T)

        T = torch.einsum('xabcdx->abcd', (T)) #trace 
        #T = torch.einsum('xabcdy,yefghx->aebfcgdh', (T, T)).contiguous().view(self.D**2, self.D**2, self.D**2, self.D**2)

        f = T.abs().max()
        lnZ += torch.log(f)/2.0**self.nxtrg
        T = T / f

        lnZ += ctmrg(T, self.chi, self.nctmrg)/2.0**self.nxtrg

        return lnZ

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser(description='')
    
    parser.add_argument("-beta", type=float, default=0.21, help="beta")
    parser.add_argument("-D", type=int, default=4, help="D")
    parser.add_argument("-chi", type=int, default=50, help="chi")
    parser.add_argument("-nxtrg", type=int, default=20, help="nxtrg")
    parser.add_argument("-nctmrg", type=int, default=50, help="nctmrg")

    parser.add_argument("-nsweeps", type=int, default=20, help="nsweeps")
    parser.add_argument("-nstays", type=int, default=1, help="nstays")
    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="GPU #")
    parser.add_argument("-use_checkpoint", action='store_true', help="use checkpoint")
    args = parser.parse_args()

    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64
    print ('use', dtype)
    print ('use checkpoint', args.use_checkpoint)

    #dimer covering
    T = torch.zeros(2, 2, 2, 2, 2, 2, dtype=dtype, device=device)
    T[0, 0, 0, 0, 0, 1] = 1.0 
    T[0, 0, 0, 0, 1, 0] = 1.0 
    T[0, 0, 0, 1, 0, 0] = 1.0 
    T[0, 0, 1, 0, 0, 0] = 1.0 
    T[0, 1, 0, 0, 0, 0] = 1.0 
    T[1, 0, 0, 0, 0, 0] = 1.0 
    
    #3d Ising
    #K = torch.tensor([args.beta], dtype=dtype, device=device)
    #c = torch.sqrt(torch.cosh(K))
    #s = torch.sqrt(torch.sinh(K))
    #Q = [c, s, c, -s]
    #Q = torch.stack(Q).view(2, 2)
    #T = torch.einsum('ai,aj,ak,al,am,an->ijklmn', [Q, Q, Q, Q, Q, Q])
   
    model = Ising(args.D, args.chi, args.nxtrg, args.nctmrg, dtype=dtype, device=device)

    def turn_on_grad(i):
        for j, p in enumerate(model.parameters()):
            if (j==i):
                p.requires_grad = True
            else:
                p.requires_grad = False

    sweeps = 0; step = 0; direction = 1
    for _ in range(args.nsweeps*args.nxtrg):
        #sweep back and forth through isometries
        turn_on_grad(step)

        for s in range(args.nstays): # local sweeps
            model.zero_grad()
            start = time.time()
            lnZ = model.forward(T) 
            forward = time.time() 
            lnZ.backward()
            backward = time.time()
            
            with torch.no_grad(): 
                #compute environment from gradient  
                #\partial tr(wE)/ \partial w = E^T 
                #cf http://www.matrixcalculus.org
                E = model.params[step].grad
                chi, chi_new = E.shape[0], E.shape[-1]
                E = E.reshape(chi**2, chi_new)
                #and perform mera update
                U, S, V = torch.svd(E)
                svd = time.time()
                
                model.params[step].data = (U@V.t()).view(chi, chi, chi_new)

        #print(forward-start, backward - forward, svd-backward)
        step += direction

        #when reaching the end reverse sweep direction and report results
        if (step==args.nxtrg-1 or step==0):
            direction *= -1
            sweeps += 1
            with torch.no_grad():
                lnZ = model.forward(T)
            print (sweeps, lnZ.item())
